module Main where

import Othello.GameLogic (GameSetup (..), Player (..), initialGameState, Coordinate, DiscState (..), movesPossibleOnBoard, playerWithBestScore, movesForPlayer, Board (..), comprehensionByBoard, opposingPlayer, applyMove, getMovesOnPoint, getAIMove)
import Graphics.Gloss
 -- We need access to events, so we use the game mode.
import Graphics.Gloss.Interface.Pure.Game
import GHC.Arr (assocs)

-- The resolution of the window
resolutionX :: Int
resolutionX = 800
resolutionY :: Int
resolutionY = 800

-- As the rendering is from the centre, in what way the coordinates should be translated to return them back to left lower edge-based positioning
centreAdjustmentX :: Int
centreAdjustmentX = -1 * (resolutionX `div` 2)
centreAdjustmentY :: Int
centreAdjustmentY = -1 * (resolutionY `div` 2)

gridAbsoluteLeftX :: Int
gridAbsoluteLeftX = 100 -- From left edge
gridAbsoluteLeftY :: Int
gridAbsoluteLeftY = 100 -- From bottom
gridBoxSize :: Int
gridBoxSize = 75 -- How large a single square is?

circleSize :: Float
 -- Radius of the circle
circleSize = 20

-- From which coordinate the text is rendered, X
textBottomLeftX :: Float
textBottomLeftX = 50
-- From which coordinate the text is rendered, Y 
textBottomLeftY :: Float
textBottomLeftY = 100

-- Assumed grid size for UI and state construction purposes
gameGridSize :: Int
gameGridSize = 8

data GameWorld = World {
    worldSetup :: GameSetup, -- Current game state
    playerTurn :: Player, -- Whose turn it is
    passedOnLastTurn :: Bool, -- Was there a pass on the last turn?
    bothStalled :: Bool, -- Has the game stalled, AKA two passes in a row, meaning game over
    ticks :: Float -- Ticks counter to count how long to wait until AI kicks into action
}

-- Declare an initial world, starting on Red, no passes or stall and starting at zero ticks.
initialWorld = World (initialGameState gameGridSize [Blue] 4) Red False False 0.0

-- Tries to locate the coordinates the mouse did click. If available, return Just it, otherwise Nothing
getClickTarget :: (Float, Float) -> Maybe Coordinate
getClickTarget (clickX, clickY)
    | dividedX < 0 || dividedX >= gameGridSize = Nothing -- Invalid coordinates
    | dividedY < 0 || dividedY >= gameGridSize = Nothing

    | otherwise = Just (dividedX, dividedY)

    where
        -- As GLOSS uses center-based coordinates, we need to translate them to start from left, bottom origin
        translatedClickX :: Int
        translatedClickX = round clickX - centreAdjustmentX - gridAbsoluteLeftX
        translatedClickY :: Int
        translatedClickY = round clickY - centreAdjustmentY - gridAbsoluteLeftY

        -- dividedX, dividedY should directly correspond to grid coordinates
        dividedX :: Int
        -- Div is an integer division; Haskell is remarkably strict about types, so we need to explicitly accept the loss of precision associated
        dividedX = translatedClickX `div` gridBoxSize

        dividedY :: Int
        dividedY = gameGridSize-1 - translatedClickY `div` gridBoxSize


-- Rendering is a bit tricky, as the render is centered to the center of the window and not to the sides! Adjustment calculations need to be made
renderWorld :: GameWorld -> Picture
renderWorld world = Translate (fromIntegral centreAdjustmentX) (fromIntegral centreAdjustmentY) (Pictures (concat [boardRender, gridList, [textRender]])) where

    playerWithBestScoreText :: Maybe Player -> String
    playerWithBestScoreText (Just Red) = "Red wins"
    playerWithBestScoreText (Just Blue) = "Blue wins"
    playerWithBestScoreText Nothing = "Draw"

    -- Text render, for status text
    textRender = Scale 0.15 0.15 (Translate textBottomLeftX textBottomLeftY $ Color white (Text (
                case () of _
                            | not (movesPossibleOnBoard (board $ worldSetup world)) -> playerWithBestScoreText (playerWithBestScore (board $ worldSetup world))
                            | null (movesForPlayer (board $ worldSetup world) (playerTurn world)) -> "No possible positions for you, " ++ show (playerTurn world) ++ ", passing"
                            | playerTurn world == Blue -> "Turn for Blue"
                            | playerTurn world == Red -> "Turn for Red"
                            | otherwise -> ""
                )))


    -- Board state, essentially grid
    boardRender = [uncurry Translate (gridPosForAssoc placedButton) $ Color (colorForAssoc placedButton) (Circle circleSize) | placedButton <- placedButtons]
        where
            gridPosForAssoc :: (Coordinate, DiscState) -> (Float, Float)
            gridPosForAssoc assoc = (actualX, actualY) where
                 (coordX, coordY) = fst assoc

                 actualX :: Float
                 actualX = fromIntegral $ gridAbsoluteLeftX + (coordX*gridBoxSize + gridBoxSize `div` 2)
                 actualY :: Float
                 actualY = fromIntegral $ gridAbsoluteLeftY + ((gameGridSize-1-coordY)*gridBoxSize + gridBoxSize `div` 2)


            colorForAssoc :: (Coordinate, DiscState) -> Color
            colorForAssoc assoc = if snd assoc == Placed Red then red else blue
            placedButtons = filter (\asc -> snd asc /= Empty) (assocs (grid $ board $ worldSetup world))


    -- List comprehension to form the drawings for the grid
    gridDrawingFunction (gridX, gridY) = whiteBox (fromIntegral $ gridAbsoluteLeftX+gridBoxSize*gridX,fromIntegral $ gridAbsoluteLeftY + gridBoxSize*gridY) (fromIntegral $ gridAbsoluteLeftX+gridBoxSize*(gridX+1),fromIntegral $ gridAbsoluteLeftY+gridBoxSize*(gridY+1))
    gridList = (comprehensionByBoard $ board $ worldSetup world) gridDrawingFunction

    -- A helper function for rendering a white box
    whiteBox startPoint endPoint = Color white (Line points) where
        (endX, endY) = endPoint
        (startX, startY) = startPoint
        points = [(startX, startY), (endX, startY), (endX, endY), (startX, endY), (startX, startY)]

-- Handle events taking place on the game world; this is essentially mouse handling only here
handleEvent :: Event -> GameWorld -> GameWorld
handleEvent (EventKey (MouseButton RightButton) Down _ _) _ = initialWorld
handleEvent (EventKey (MouseButton LeftButton) Down _ clickPos) world
    -- Both players have passed, so the game's over
    | bothStalled world = world
    --  AI plays this turn
    | playerTurn world `elem` aiPlays (worldSetup world) = world
    -- No valid position
    | Nothing <- possibleClickPos = world
    -- We have a position,  evaluate it
    | Just coordinate <- possibleClickPos = evaluatePlayerTurn world coordinate
    where
        evaluatePlayerTurn :: GameWorld -> Coordinate -> GameWorld
        evaluatePlayerTurn world coord
            -- No valid moves
            | null moveOnPoint = world
            -- Apply a move and change the turn to the opposing player; also reset any pass counters and the ticker
            | otherwise = World (setup {board = appliedBoard}) (opposingPlayer (playerTurn world)) False False 0
            where
                setup = worldSetup world
                appliedBoard = applyMove (board $ worldSetup world) (playerTurn world) moveOnPoint
                moveOnPoint = getMovesOnPoint (board $ worldSetup world) (playerTurn world) coord
        possibleClickPos = getClickTarget clickPos
-- Rest do not affect the world
handleEvent _ world = world

-- Timer ticks, some actions cause an intentional delay which is handled here
timerTick :: Float -> GameWorld -> GameWorld
timerTick tick_diff world@(World setup turn passedOnLast bothStalled curTicks)
    | tick_diff+curTicks < 1.0 = world { ticks = curTicks+tick_diff}
    | otherwise = evaluateTickTurn
    where
        evaluateTickTurn
          -- Both have stalled, do not do anything
          | bothStalled = tickResetWorld
          -- Pass, unable to make a turn
          | null (movesForPlayer (board setup) turn) = World setup (opposingPlayer turn) True passedOnLast 0
          -- AI does not play this turn
          | turn `notElem` aiPlays setup = tickResetWorld
          -- AI plays this turn and resets the ticks
          | otherwise = World (setup {board = applyMove (board setup) turn aiTurn}) (opposingPlayer turn) False False 0

        -- No changes apart from the ticks resetting to zero, adding a delay
        tickResetWorld = world { ticks = 0 }

        -- AI turn
        aiTurn = getAIMove setup turn

main = play -- This program is a GLOSS game..
        (InWindow "Othello Haskell" (resolutionX,resolutionY) (20,20)) -- In a window of a suitable size
         black -- With a black background
         24 -- 24fps
         initialWorld -- Initial state
         renderWorld -- Render images using renderWorld
         handleEvent -- Handle events using handleEvent
         timerTick -- Timer effects via timerTick